#!/usr/bin/env python3

cmd_desc = "Translate MIDI data streams and files into OSC messages to drive the system."

import argparse
import time
import logging

# Mido 'MIDI objects' package
# https://mido.readthedocs.io/en/latest/index.html
import mido

# common MIDI event processing
from theater.midi import MIDItoOSC

# common logging functions
import theater.logging

# current setup
import theater.config

# initialize logging for this module
log = logging.getLogger('bridge')

#================================================================
class EventLoop:
    def __init__(self, args):

        # Open the MIDI input port
        names = mido.get_input_names()
        if args.midi in names:
            self.midi_port = mido.open_input(args.midi)
        else:
            log.warning("Requested MIDI input %s not available in %s", args.midi, names)
            self.midi_port = mido.open_input(names[0])
        log.info("Opened MIDI input port %s", self.midi_port.name)

        # Create the networking output.
        self.bridge = MIDItoOSC(args)

        return

    def close(self):
        self.midi_port.close()
        self.bridge.close()

    #---------------------------------------------------------------
    def run(self):
        while True:
            msg = self.midi_port.receive()
            self.bridge.perform_event(msg)

#================================================================
# The following section is run when this is loaded as a script.
if __name__ == "__main__":

    # set up logging
    theater.logging.open_log_file('logs/MIDI-bridge.log')
    log.info("Starting MIDI-bridge.py")

    # Initialize the command parser.
    parser = argparse.ArgumentParser(description = cmd_desc)
    parser.add_argument( '--midi', default='IAC Driver Bus 1', help='MIDI input device (default: %(default)s).')
    parser.add_argument("--ip", default=theater.config.theater_IP,  help="IP address of the OSC receivers (default: %(default)s).")
    theater.logging.add_logging_args(parser)

    # Parse the command line, returning a Namespace.
    args = parser.parse_args()

    # Modify logging settings as per common arguments.
    theater.logging.configure_logging(args)

    # Create the event loop.
    event_loop = EventLoop(args)

    # Begin processing MIDI input.  This may be safely interrupted by the user pressing Control-C.
    try:
        event_loop.run()

    except KeyboardInterrupt:
        log.info("User interrupted operation.")
        print("User interrupt, shutting down.")
        event_loop.close()

    log.info("Exiting MIDI-bridge.py")
