#!/usr/bin/env python3

cmd_desc = "Replay a MIDI file in real time as system commands."

import argparse
import time
import logging

# Mido 'MIDI objects' package
# https://mido.readthedocs.io/en/latest/index.html
import mido

# common MIDI event processing
from theater.midi import MIDItoOSC

# common logging functions
import theater.logging

# library schedule
import theater.HuntLibrary as HuntLibrary

# initialize logging for this module
log = logging.getLogger('player')

#================================================================

class Player:
    """MIDI file player."""

    def __init__(self, args):

        self.verbose = args.verbose

        self.midifile = mido.MidiFile(args.input)
        self.log_midifile_metadata()

        # MIDI files express tempo as microseconds per quarter note; this sets
        # a default tempo of 120 BPM (0.5 sec per beat).
        self.midi_tempo = 500000

        # MIDI files express times in integer ticks per quarter note.
        self.ticks_per_beat = self.midifile.ticks_per_beat

        # Merge all tracks (if more than one) into a single track in time order.
        # This assumes that events are identified by channel number for routing
        # and not simply by track.
        self.playtrack = mido.merge_tracks(self.midifile.tracks)

        # Create the networking output.
        self.bridge = MIDItoOSC(args)

        return

    def close(self):
        self.bridge.close()

    #--------------------------------------------------------------------
    def log_midifile_metadata(self):
        log.info("Opened MIDI file %s, MIDI format %d, %d tracks, %d ticks per beat",
                 args.input, self.midifile.type, len(self.midifile.tracks), self.midifile.ticks_per_beat)

        for i, track in enumerate(self.midifile.tracks):
            log.info("MIDI file track %d: %s", i, track.name)

        # report some diagnostics on the current midifile to the log
        for track in self.midifile.tracks:
            log.info("Contents of track '%s'", track.name)
            event_messages = 0
            for msg in track:
                if msg.is_meta:
                    log.info("MIDI metadata: %s", msg)
                else:
                    event_messages += 1
            # end of track
            log.info("Track '%s' includes %d events.", track.name, event_messages)

    #--------------------------------------------------------------------
    def perform_event(self, event):
        # ignore metadata messages track_name, instrument_name, key_signature, smpte_offset, etc.
        if event.type == 'set_tempo':
            log.info("MIDI tempo change: %d usec/beat (%f BPM).", event.tempo, mido.tempo2bpm(event.tempo))
            self.midi_tempo = event.tempo

        elif event.type in ['note_on', 'note_off']:
            log.debug("Note On: channel %d, note %d, velocity %d", event.channel, event.note, event.velocity)
            self.bridge.perform_event(event)

    #--------------------------------------------------------------------
    def run(self, skip=0):
        """Run one performance of the MIDI file and return.  Eeach MIDI event is
        issued in real time as OSC network messages translated via the MIDI to
        OSC bridge object."""

        # Keep track of target times in integer nanoseconds to avoid roundoff
        # errors.  Following an absolute clock will maintain overall precision
        # in the presence of sleep time jitter.
        start_t = time.monotonic_ns()
        next_timepoint = start_t

        # play through all the events in the sequence, waiting the specified
        # number of ticks before each issue
        for event in self.playtrack:

            if self.verbose:
                log.debug("Playing event: %s", event)

            # don't perform the end of track, it can have an unreasonable delay
            if event.type == 'end_of_track':
                log.info("Found end event: %s", event)
                return

            if skip > 0:
                skip -= 1

            else:
                # if the next event has a preceding delay, convert from ticks
                # at the current tempo to an absolute time in nanoseconds
                if event.time > 0:
                    event_ns = event.time * 1000 * self.midi_tempo / self.ticks_per_beat
                    next_timepoint += event_ns
                    delay_ns = next_timepoint - time.monotonic_ns()
                    if delay_ns > 0:
                        time.sleep(delay_ns * 1e-9)

                self.perform_event(event)

#================================================================
# The following section is run when this is loaded as a script.
if __name__ == "__main__":

    # set up logging
    theater.logging.open_log_file('logs/MIDI-player.log')
    log.info("Starting MIDI-player.py")

    # Initialize the command parser.
    parser = argparse.ArgumentParser(description = cmd_desc)
    parser.add_argument( '--ip', default="127.0.0.1",  help="IP address of the OSC receivers (default: %(default)s).")
    parser.add_argument( '--scheduled',action='store_true',  help="Enable continuous operation on schedule.")
    parser.add_argument( '--skip',type=int, default=0,  help="Specify number of events to skip.")
    theater.logging.add_logging_args(parser)
    parser.add_argument( 'input', help='MIDI file to perform.')

    # Parse the command line, returning a Namespace.
    args = parser.parse_args()

    # Modify logging settings as per common arguments.
    theater.logging.configure_logging(args)

    # Create the MIDI event player.
    player = Player(args)

    # Run the performance until done. This may be safely interrupted by the user pressing Control-C.
    try:

        if args.scheduled:
            # run the file repeatedly whenever the scheduler indicates show time
            while True:
                if HuntLibrary.current_time_properties().get('is_show_time'):
                    log.info("Starting scheduled iteration.")
                    player.run()
                    log.info("Starting brief intermission pause.")
                    time.sleep(5)
                else:
                    time.sleep(60) # wait before checking time again
        else:
            log.info("Starting a single performance.")
            player.run(args.skip)

    except KeyboardInterrupt:
        log.info("User interrupted operation.")
        print("User interrupt, shutting down.")
        player.close()

    log.info("Exiting MIDI-player.py")
