#!/usr/bin/env python3

cmd_desc = "Show controller script to start and stop individual performances."

import argparse
import time
import math
import logging
import threading
import subprocess
import signal

import numpy as np

# import the pythonosc package
from pythonosc import dispatcher
from pythonosc import osc_server

# OSC port for lighting commands
import theater.config
import theater.network

# library schedule
import theater.HuntLibrary as HuntLibrary

# import common logging functions
import theater.logging

# initialize logging for this module
log = logging.getLogger('director')

#================================================================
class Director:
    def __init__(self, args):

        # Start without any child performance process.
        self.performance = None

        # Create the networking listener
        self.init_networking_input(args)

        # Create the networking senders
        self.network = theater.network.TheaterNetwork(args)
        return

    #---------------------------------------------------------------
    def close(self):
        log.info("Closing ports.")
        self.server.shutdown()

    #---------------------------------------------------------------
    def init_networking_input(self, args):

        # Initialize the OSC message dispatch system.
        self.dispatch = dispatcher.Dispatcher()
        self.dispatch.map("/audience",   self.audience)
        self.dispatch.set_default_handler(self.unknown_message)

        # Start and run the server.
        udp_port = theater.config.director_UDP_port
        listen_IP = '0.0.0.0'
        self.server = osc_server.OSCUDPServer((listen_IP, udp_port), self.dispatch)
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.daemon = True
        self.server_thread.start()
        log.info("started OSC server on port %s:%d", listen_IP, udp_port)

    def unknown_message(self, msgaddr, *args):
        """Default handler for unrecognized OSC messages."""
        log.warning("Received unmapped OSC message %s: %s", msgaddr, args)

    def audience(self, msgaddr, *args):
        """Process /audience message received via OSC over UDP."""
        # log.debug("received OSC message %s: %s", msgaddr, args)
        if self.performance is None:
            log.debug("Running idle performance.")
            self.performance = subprocess.Popen(["./idle-show.py", "--debug"])

    #---------------------------------------------------------------
    def start_show(self):
        if self.performance is None:
            log.debug("Running demo performance.")
            self.performance = subprocess.Popen(["./demo-show.py", "--debug"])

    #---------------------------------------------------------------
    # Event loop.
    def run(self):
        """Run a performance event loop to execute periodic state updates."""

        while True:
            if self.performance is not None:
                returncode = self.performance.poll()
                if returncode is None:
                    log.debug("Waiting for performance to finish.")
                    time.sleep(5)

                else:
                    log.debug("Performance finished.")
                    self.performance = None

            else:
                log.debug("Sitting idle.")
                time.sleep(10)

                if HuntLibrary.current_time_properties().get('is_show_time'):
                    log.info("Starting scheduled iteration.")
                    self.start_show()

#================================================================
# The following section is run when this is loaded as a script.
if __name__ == "__main__":

    # set up logging
    theater.logging.open_log_file('logs/director.log')
    log.info("Starting director.py")
    np.set_printoptions(linewidth=np.inf)

    # Initialize the command parser.
    parser = argparse.ArgumentParser(description = cmd_desc)
    parser.add_argument("--ip", default=theater.config.theater_IP,  help="IP address of the OSC receivers (default: %(default)s).")
    theater.logging.add_logging_args(parser)

    # Parse the command line, returning a Namespace.
    args = parser.parse_args()

    # Modify logging settings as per common arguments.
    theater.logging.configure_logging(args)

    # Create the performance controller.
    director = Director(args)

    # Begin the performance.  This may be safely interrupted by the user pressing Control-C.
    try:
        director.run()

    except KeyboardInterrupt:
        log.info("User interrupted operation.")
        print("User interrupt, shutting down.")
        director.close()

    log.info("Exiting director.py")
