#!/usr/bin/env python3
# Obtain and process images from a Blackfly GigE machine vision camera using the FLIR Spinnaker PySpin API.

cmd_desc = "Detect hallway motion using a FLIR camera."

import os
import sys
import logging
import argparse
import datetime

# OpenCV
import cv2 as cv

# NumPy
import numpy as np

# FLIR Spinnaker camera API
import PySpin

# common logging functions
import theater.logging

# communicate to other nodes
import theater.network

# initialize logging for this module
log = logging.getLogger('vision')

#================================================================
class SpinnakerCamera:
    """Representation of a camera using the FLIR Spinnaker PySpin API."""

    #----------------------------------------------------------------
    def start_continuous(self):
        #  Retrieve enumeration node from nodemap
        # In order to access the node entries, they have to be casted to a pointer type (CEnumerationPtr here)
        node_acquisition_mode = PySpin.CEnumerationPtr(self.nodemap.GetNode('AcquisitionMode'))
        if not PySpin.IsReadable(node_acquisition_mode) or not PySpin.IsWritable(node_acquisition_mode):
            log.error("Unable to set acquisition mode to continuous (enum retrieval).")
            return

        # Retrieve entry node from enumeration node
        node_acquisition_mode_continuous = node_acquisition_mode.GetEntryByName('Continuous')
        if not PySpin.IsReadable(node_acquisition_mode_continuous):
            log.error("Unable to set acquisition mode to continuous (entry retrieval)")
            return

        # Retrieve integer value from entry node
        acquisition_mode_continuous = node_acquisition_mode_continuous.GetValue()

        # Set integer value from entry node as new value of enumeration node
        node_acquisition_mode.SetIntValue(acquisition_mode_continuous)
        self.continuous_mode = True

        log.info("Acquisition mode set to continuous.")
        self.cam.BeginAcquisition()

        # Create ImageProcessor instance for post processing images
        self.processor = PySpin.ImageProcessor()

        # Set default image processor color processing method
        self.processor.SetColorProcessing(PySpin.SPINNAKER_COLOR_PROCESSING_ALGORITHM_HQ_LINEAR)

        return

    #----------------------------------------------------------------
    def end_continuous(self):
        if self.continuous_mode:
            log.info("Ending continuous acquisition mode.")
            self.cam.EndAcquisition()
            self.continuous_mode = False

    #----------------------------------------------------------------
    def acquire_image(self):
        image_result = self.cam.GetNextImage(1000)

        if image_result.IsIncomplete():
            if self.args.verbose:
                log.debug("Image incomplete with image status %d", image_result.GetImageStatus())
            return False

        else:
            width = image_result.GetWidth()
            height = image_result.GetHeight()

            if self.args.verbose:
                log.debug("Grabbed with width = %d, height = %d", width, height)

            #  Convert image to mono 8
            # The result is an instance of PySpin.ImagePtr
            image_converted = self.processor.Convert(image_result, PySpin.PixelFormat_Mono8)

            # Retrieve image object compatible with OpenCV
            image = image_converted.GetData().reshape(height, width, 1)
            result = image

            #  Release image
            image_result.Release()

            return result


    #================================================================
    def log_device_info(self):
        log.info("*** DEVICE INFORMATION ***")
        node_device_information = PySpin.CCategoryPtr(self.nodemap_tldevice.GetNode('DeviceInformation'))

        if PySpin.IsReadable(node_device_information):
            features = node_device_information.GetFeatures()
            for feature in features:
                node_feature = PySpin.CValuePtr(feature)
                log.info('%s: %s', node_feature.GetName(),
                                  node_feature.ToString() if PySpin.IsReadable(node_feature) else 'Node not readable')
        else:
            log.warning("Device control information not readable.")

    #================================================================
    def __init__(self, args):

        self.args = args
        self.continuous_mode = False

        # Retrieve singleton reference to system object
        self.system = PySpin.System.GetInstance()

        # Get current library version
        version = self.system.GetLibraryVersion()
        log.info('PySpin library version: %d.%d.%d.%d', version.major, version.minor, version.type, version.build)

        # Retrieve list of cameras from the system
        self.cam_list = self.system.GetCameras()
        num_cameras = self.cam_list.GetSize()
        log.info("Number of cameras detected: %d", num_cameras)

        if num_cameras == 0:
            print("No cameras found.")
            self.cam = None

        elif num_cameras > 1:
            print("Warning, detected more than one camera, only using first one.")
            self.cam = self.cam_list[0]

        else:
            self.cam = self.cam_list[0]

        if self.cam is not None:
            # Retrieve transport layer device nodemap
            self.nodemap_tldevice = self.cam.GetTLDeviceNodeMap()

            # Optionally log device information
            self.log_device_info()

            # Initialize camera
            self.cam.Init()

            # Retrieve GenICam nodemap
            self.nodemap = self.cam.GetNodeMap()

    #----------------------------------------------------------------
    def close(self):
        log.debug("SpinnakerCamera close called.")

        # End acquisition if active
        self.end_continuous()

        # Deinitialize camera
        self.cam.DeInit()

        # Release reference to camera
        del self.cam

        # Clear camera list before releasing system
        self.cam_list.Clear()

        # Release system instance
        self.system.ReleaseInstance()

    def __del__(self):
        log.debug("SpinnakerCamera __del__ called.")
        self.close()

#================================================================
class EventLoop:
    def __init__(self, camera, args):

        self.camera = camera
        self.args = args
        self.last_capture = None

        # load the current camera image maask
        self.motion_mask = cv.imread("images/hallway-mask.png", 0)
        value, self.motion_mask = cv.threshold(self.motion_mask, 127, 255, cv.THRESH_BINARY)

        # create a background subtractor
        self.background_subtractor = cv.createBackgroundSubtractorMOG2()

        # Create the networking output.
        self.network = theater.network.TheaterNetwork(args)

        return

    #---------------------------------------------------------------
    def is_motion_detected(self, image):
        ret, thresholded = cv.threshold(image, 127, 255, cv.THRESH_BINARY)
        dilated = cv.dilate(thresholded, None, iterations=0)

        # if self.args.viewer:
        #    cv.imshow("Debug", dilated)

        contours, hierarchy = cv.findContours(dilated, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)
        for contour in contours:
            area = cv.contourArea(contour)
            if area > 200:
                log.debug("Found image difference area %f", area)
                return True

        return False

    #---------------------------------------------------------------
    def write_capture_file(self, image):

        now = datetime.datetime.now()
        if self.last_capture is None or (now - self.last_capture).total_seconds() > 10:
            folder = os.path.expanduser("~/tmp/captures")
            filename = now.strftime("%Y-%m-%d-%H-%M-%S.png")
            path = os.path.join(folder, filename)
            cv.imwrite(path, image)
            self.last_capture = now
            log.debug("Wrote image to %s", path)
        else:
            log.debug("Skipping image capture, too soon.")

    #---------------------------------------------------------------
    def run(self):
        while True:
            image = self.camera.acquire_image()

            # ignore empty images
            if image is False:
                continue

            # immediately subsample
            image = image[::3, ::3]
            subsampled = image

            # show the source
            if self.args.viewer:
                cv.imshow("Source", image)

            # crop a portion of the image for motion processing
            if self.args.crop:
                image = cv.bitwise_and(image, self.motion_mask)

            if self.args.background:
                image = self.background_subtractor.apply(image)

            detected = self.is_motion_detected(image)
            if detected:
                self.write_capture_file(subsampled)
                self.network.director.send_message("/audience", [])

            # optionally display the processed results
            if self.args.viewer:
                cv.imshow("Processed", image)

                # Calling waitKey allows the window system event loop to run and update
                # the display.  The waitKey argument is in milliseconds.
                if cv.waitKey(33) & 0xFF == ord('q'):
                    break

#================================================================
if __name__ == '__main__':

    # set up logging
    theater.logging.open_log_file('logs/hallway-monitor.log')
    log.info("Starting hallway-monitor.py")

    # Initialize the command parser.
    parser = argparse.ArgumentParser(description = cmd_desc)
    parser.add_argument( '--ip', default="127.0.0.1",  help="IP address of the OSC receivers (default: %(default)s).")
    parser.add_argument( '--viewer', action='store_true', help="Enable GUI viewer window.")
    parser.add_argument( '--crop', action='store_false', help="Disable image cropping.")
    parser.add_argument( '--background', action='store_false', help='Disable background subtraction.')

    theater.logging.add_logging_args(parser)

    # Parse the command line, returning a Namespace.
    args = parser.parse_args()

    # Modify logging settings as per common arguments.
    theater.logging.configure_logging(args)

    # Open the camera.
    camera = SpinnakerCamera(args)
    camera.start_continuous()

    if not camera.continuous_mode:
        log.error("Camera not acquiring, quitting.")

    else:
        if args.viewer:
            cv.namedWindow("Source")
            cv.namedWindow("Processed")
            # cv.namedWindow("Debug")

        # Create the event loop.
        event_loop = EventLoop(camera, args)

        try:
            event_loop.run()

        except KeyboardInterrupt:
            print ("User quit.")

    log.info("Exiting hallway-monitor.py")
